"""Main entry point for the service."""
import os
from os import path
import logging.config
import yaml

from rest_model_service.configuration import ServiceConfiguration
from rest_model_service.helpers import create_app

try:
    from prometheus_fastapi_instrumentator import Instrumentator
except ImportError:
    Instrumentator = None


logger = logging.getLogger(__name__)


# getting the path to the configuration file
# defaults to "rest_config.yaml" in the current directory if the path not provided
if "REST_CONFIG" in os.environ:
    file_path = os.environ["REST_CONFIG"]
else:
    file_path = "rest_config.yaml"

# loading the configuration
if path.exists(file_path) and path.isfile(file_path):
    with open(file_path) as file:
        configuration_dict = yaml.full_load(file)

    # loading configuration
    configuration = ServiceConfiguration(**configuration_dict)

    # creating app
    app = create_app(configuration, wait_for_model_creation=False)

    # instrumenting the app if metrics are enabled
    if Instrumentator is None and configuration.metrics is not None and configuration.metrics.enabled:
        logger.critical("Cannot enable metrics because optional dependency 'metrics' is not installed.")
        raise RuntimeError("Cannot start service with metrics without optional dependencies.")

    if configuration.metrics is not None and configuration.metrics.enabled:
        Instrumentator(**configuration.metrics.model_dump(exclude=["enabled"])).instrument(app).expose(
            app, include_in_schema=False)
else:
    # if there is no configuration file, or it is not found, then raise an exception
    raise ValueError("Could not find configuration file '{}', service has no models loaded.".format(file_path))



"""Configuration for the service."""
from typing import List, Dict, Any, Optional
from pydantic import BaseModel, Field


class ModelDecorator(BaseModel):
    """Settings for a decorator for a model instance in the service."""

    class_path: str = Field(description="Class path of the decorator class.")
    configuration: Optional[Dict[str, Any]] = Field(default=None, description="Configuration to initialize decorator "
                                                                              "instance.")


class MetricsConfiguration(BaseModel):
    """Configuration for metrics."""

    enabled: bool = Field(default=False, description="Enable metrics.")
    should_group_status_codes: bool = Field(default=True, description="Should status codes be grouped into `2xx`, "
                                                                      "`3xx` and so on?")
    should_ignore_untemplated: bool = Field(default=False, description="Should requests without a matching template be "
                                                                       "ignored?")
    should_group_untemplated: bool = Field(default=True, description="Should requests without a matching template "
                                                                     "be grouped to handler `none`?")
    should_round_latency_decimals: bool = Field(default=False, description="Should recorded latencies be rounded to a "
                                                                           "certain number of decimals?")
    should_respect_env_var: bool = Field(default=False, description="Should the instrumentator only work if a certain "
                                                                    "environment variable is set to `true`?")
    should_instrument_requests_inprogress: bool = Field(default=False, description="Enables a gauge that shows the "
                                                                                   "inprogress requests.")
    excluded_handlers: List[str] = Field(default=[], description="List of strings that will be compiled to regex "
                                                                 "patterns. All matches will be skipped and not "
                                                                 "instrumented.")
    body_handlers: List[str] = Field(default=[], description="List of strings that will be compiled to regex patterns. "
                                                             "All matches will be instrumented and the body will be "
                                                             "parsed.")
    round_latency_decimals: int = Field(default=4, description="Number of decimals latencies should be rounded to.")

    env_var_name: str = Field(default="ENABLE_METRICS", description="Any valid os environment variable name that will "
                                                                    "be checked for existence before instrumentation.")
    inprogress_name: str = Field(default="http_requests_inprogress",
                                 description="Name of the gauge. Defaults to "
                                             "`http_requests_inprogress`. Ignored unless "
                                             "`should_instrument_requests_inprogress` is `True`.")
    inprogress_labels: bool = Field(default=False, description="Should labels `method` and `handler` be part of the "
                                                               "inprogress label? Ignored unless "
                                                               "`should_instrument_requests_inprogress` is `True`. "
                                                               "Defaults to `False`.")


class Model(BaseModel):
    """Settings for a single model in the service."""

    class_path: str = Field(description="Class path of the model class.")
    create_endpoint: bool = Field(description="Whether or not to create an endpoint for the model.")
    decorators: Optional[List[ModelDecorator]] = Field(default=None, description="List of decorators to attach to "
                                                                                 "model.")
    configuration: Optional[Dict[str, Any]] = Field(default=None, description="Configuration to initialize model "
                                                                              "instance.")


class ServiceConfiguration(BaseModel):
    """Configuration for the service."""

    service_title: str = "RESTful Model Service"
    version: str = "0.1.0"
    description: str = ""
    models: List[Model] = Field(default=[], description="Model configuration.")
    logging: Optional[Dict] = Field(default=None, description="Logging configuration.")
    metrics: Optional[MetricsConfiguration] = Field(default=None, description="Metrics configuration.")



"""Exception handler."""
import logging
from starlette.responses import JSONResponse
from fastapi import Request
from fastapi.exceptions import RequestValidationError

from rest_model_service.schemas import Error


async def validation_exception_handler(request: Request, exception: RequestValidationError) -> JSONResponse:
    """Exception handler."""
    logger = logging.getLogger(__name__)

    messages = []
    for error in exception.errors():
        if type(error) is dict:
            message = "Field '{}' has error '{}', {}.".format(", ".join(error["loc"]),
                                                              error["type"],
                                                              error["msg"])
            messages.append(message)
        else:
            messages.append(str(error))

    extra = {
        "action": "predict",
        "endpoint": str(request.url),
        "status": "error",
        "error_info": ", ".join(messages)
    }
    logger.error(msg="Request validation error.", extra=extra)

    error = Error(type="ValidationError", messages=messages).model_dump()
    return JSONResponse(status_code=400, content=error)
